import pytest
import numpy as np
import proper_tea as pt
import proper_tea.numpy


@pytest.fixture
def numpy_array_test_class():
    class MyClass:
        any_array = pt.numpy.numpy_array()
        int_array = pt.numpy.numpy_array(dtype=int)
        float_array = pt.numpy.numpy_array(dtype=float)
        array_1D = pt.numpy.numpy_array(shape=3)
        array_2D = pt.numpy.numpy_array(shape=(2, 3))
        sort_array = pt.numpy.numpy_array(shape=2, dtype=float, sort=True)

    return MyClass()


def test_numpy_array_with_1D_list(numpy_array_test_class):
    test_class = numpy_array_test_class

    test_class.any_array = [1, 2, 3]
    assert isinstance(test_class.any_array, np.ndarray)
    assert test_class.any_array.ndim == 1
    assert np.all(test_class.any_array == [1, 2, 3])


def test_numpy_array_with_2D_list(numpy_array_test_class):
    test_class = numpy_array_test_class

    test_class.any_array = [[1.5, 2.1, 3.6], [4.1, 5.0, 6.9]]
    assert isinstance(test_class.any_array, np.ndarray)
    assert test_class.any_array.dtype == float
    assert test_class.any_array.ndim == 2


def test_numpy_int_array_with_2D_list(numpy_array_test_class):
    test_class = numpy_array_test_class

    test_class.int_array = [[1.5, 2.1, 3.6], [4.1, 5.0, 6.9]]
    assert isinstance(test_class.int_array, np.ndarray)
    assert test_class.int_array.dtype == int
    assert test_class.int_array.ndim == 2
    assert np.all(test_class.int_array == [[1, 2, 3], [4, 5, 6]])


def test_numpy_float_array_with_2D_list(numpy_array_test_class):
    test_class = numpy_array_test_class

    test_class.float_array = [[1, 2, 3], [4, 5, 6]]
    assert isinstance(test_class.float_array, np.ndarray)
    assert test_class.float_array.dtype == float
    assert test_class.float_array.ndim == 2
    assert np.all(test_class.float_array == [[1, 2, 3], [4, 5, 6]])


def test_numpy_array_with_numpy_input(numpy_array_test_class):
    test_class = numpy_array_test_class

    # Set with NumPy arrays
    x = np.linspace(0.0, 100.0, 1001, dtype=np.float16)

    test_class.any_array = x
    assert isinstance(test_class.any_array, np.ndarray)
    assert test_class.any_array.dtype == np.float16
    assert test_class.any_array.ndim == 1
    assert len(test_class.any_array) == 1001
    assert np.all(test_class.any_array == x)
    # It should reference the exact same array in memory
    assert test_class.any_array is x


def test_numpy_int_array_with_numpy_input(numpy_array_test_class):
    test_class = numpy_array_test_class

    # Set with NumPy arrays
    x = np.linspace(0.0, 100.0, 1001, dtype=np.float16)

    test_class.int_array = x
    assert isinstance(test_class.int_array, np.ndarray)
    assert test_class.int_array.dtype == int
    assert test_class.int_array.ndim == 1
    assert len(test_class.int_array) == 1001
    assert np.all(test_class.int_array == np.array(x, dtype=int))
    # It should NOT reference the same array in memory
    assert test_class.int_array is not x


def test_numpy_float_array_with_numpy_input(numpy_array_test_class):
    test_class = numpy_array_test_class

    # Set with NumPy arrays
    x = np.linspace(0.0, 100.0, 1001, dtype=np.float16)

    test_class.float_array = x

    assert isinstance(test_class.float_array, np.ndarray)
    assert test_class.float_array.dtype == float
    assert test_class.float_array.ndim == 1
    assert len(test_class.float_array) == 1001
    assert np.all(np.isclose(test_class.float_array, np.array(x, dtype=float)))
    # It should NOT reference the same array in memory (float != np.float16)
    assert test_class.float_array is not x


def test_numpy_array_with_numpy_input_does_not_copy(numpy_array_test_class):
    test_class = numpy_array_test_class

    # Set with NumPy arrays
    x = np.linspace(0.0, 100.0, 1001)

    # If conversion isn't required, it should not create a copy
    test_class.any_array = x
    assert test_class.any_array is x


def test_fixed_size_numpy_array_1D(numpy_array_test_class):
    test_class = numpy_array_test_class

    # Test it can accept any 3-vector type input
    test_class.array_1D = [1, 2, 3]
    assert len(test_class.array_1D) == 3
    test_class.array_1D = (4.0, 5.0, 6.0)
    assert len(test_class.array_1D) == 3
    test_class.array_1D = np.linspace(1.0, 3.0, 3)
    assert len(test_class.array_1D) == 3

    # Test errors are raised when setting incorrect shape
    with pytest.raises(ValueError) as excinfo:
        test_class.array_1D = [1, 2]
    assert "shape" in str(excinfo.value)
    with pytest.raises(ValueError) as excinfo:
        test_class.array_1D = (3,)
    assert "shape" in str(excinfo.value)
    with pytest.raises(ValueError) as excinfo:
        test_class.array_1D = 5
    assert "shape" in str(excinfo.value)
    with pytest.raises(ValueError) as excinfo:
        test_class.array_1D = np.linspace(1.0, 4.0, 4).reshape((2, 2))
    assert "shape" in str(excinfo.value)
    # should be unchanged after these attempts
    assert len(test_class.array_1D) == 3


def test_fixed_size_numpy_array_2D(numpy_array_test_class):
    test_class = numpy_array_test_class

    # Test it can accept any 3-vector type input
    test_class.array_2D = [[1, 2, 3], [4, 5, 6]]
    assert np.all(np.equal(test_class.array_2D.shape, (2, 3)))
    test_class.array_2D = ((4.0, 5.0, 6.0), (7.0, 8.0, 9.0))
    assert np.all(np.equal(test_class.array_2D.shape, (2, 3)))
    test_class.array_2D = np.linspace(1.0, 6.0, 6).reshape((2, 3))
    assert np.all(np.equal(test_class.array_2D.shape, (2, 3)))

    # Test errors are raised when setting incorrect shape
    with pytest.raises(ValueError) as excinfo:
        test_class.array_2D = [1, 2, 3, 4, 5, 6]
    assert "shape" in str(excinfo.value)
    with pytest.raises(ValueError) as excinfo:
        test_class.array_2D = (3,)
    assert "shape" in str(excinfo.value)
    with pytest.raises(ValueError) as excinfo:
        test_class.array_2D = 5
    assert "shape" in str(excinfo.value)
    with pytest.raises(ValueError) as excinfo:
        test_class.array_2D = np.linspace(1.0, 4.0, 4).reshape((2, 2))
    assert "shape" in str(excinfo.value)
    # should be unchanged after these attempts
    assert np.all(np.equal(test_class.array_2D.shape, (2, 3)))


def test_sorted_numpy_array(numpy_array_test_class):
    test_class = numpy_array_test_class

    # Test it can accept a sorted 2-vector
    test_class.sort_array = [1, 2]
    assert test_class.sort_array[0] == 1
    assert test_class.sort_array[1] == 2

    # Test it automatically sorted an unsorted 2-vector
    test_class.sort_array = [2, 1]
    assert test_class.sort_array[0] == 1
    assert test_class.sort_array[1] == 2



try:
    from importlib.metadata import version, PackageNotFoundError
except (ModuleNotFoundError, ImportError):
    from importlib_metadata import version, PackageNotFoundError
try:
    __version__ = version("proper_tea")
except PackageNotFoundError:
    from setuptools_scm import get_version

    __version__ = get_version(root="..", relative_to=__file__)

__all__ = ["__version__"]

from .property_factory import property_factory
from .constrained_numbers import (
    floating_point,
    integer,
    boolean,
    positive,
    positive_int,
    positive_float,
    negative,
    negative_int,
    negative_float,
    greater_than,
    int_greater_than,
    float_greater_than,
    less_than,
    int_less_than,
    float_less_than,
    in_range,
    int_in_range,
    float_in_range,
    not_in_range,
    int_not_in_range,
    float_not_in_range,
)
from .discrete_sets import in_set



"""property_factory

Defines a generic property factory which allows the user to set a transform
and a condition that must be met by any inputs to a property.
"""


def property_factory(
    condition=None,
    transform=None,
    condition_err_msg: str = "",
    transform_err_msg: str = "",
):
    """A generic function for creating class properties.

    This function allows the user to assert a condition that must be satisfied
    within the setter, along with the type the property should be set to. It
    does not impose any conditions on the getter.

    Parameters:

        condition : Should either be None, meaning any input is accepted, or a
            function taking one argument and returning a bool. An example of a
            common condition may be the property must be non-negative.
        transform : Should either be None, meaning the value is not modified,
            or function that takes one argument. This is used to modify inputs
            to the property. If set to a type (e.g. int, float, str), transform
            will attempt to cast the input to that type.
        condition_err_msg (str): Message to print if 'condition' returns False.
        transform_err_msg (str): Message to print if casting to 'transform'
            fails.

    Returns:

        property : A property decorator. For details on properties, see
            https://realpython.com/python-property/
    """

    # Use counter for unique underlying names
    try:
        property_factory.counter += 1
    except AttributeError:
        property_factory.counter = 0

    internal_name = f"_proper_tea:attr{property_factory.counter}"

    def getter(instance):
        """Retrieve the property from the given class instance."""
        return getattr(instance, internal_name)

    def setter(instance, value):
        """Assign 'value' to the property within the given class instance."""

        if condition is not None:
            try:
                valid = condition(value)
            except Exception as e:
                err_msg = (
                    f"Setter condition for property in class "
                    f"{instance.__class__.__name__}"
                    f"raised exception. See traceback for more info."
                )
                if condition_err_msg:
                    err_msg += f"\n{condition_err_msg}"
                raise ValueError(err_msg) from e

            if not valid:
                err_msg = (
                    f"Setter condition for property in class "
                    f"{instance.__class__.__name__} "
                    f"returned False."
                )
                if condition_err_msg:
                    err_msg += f"\n{condition_err_msg}"
                raise ValueError(err_msg)

        if transform is not None:
            try:
                value = transform(value)
            except Exception as e:
                err_msg = (
                    f"Setter transform for property in class "
                    f"{instance.__class__.__name__} "
                    f"raised exception. See traceback for more info."
                )
                if transform_err_msg:
                    err_msg += f"\n{transform_err_msg}"
                raise ValueError(err_msg) from e

        setattr(instance, internal_name, value)

    return property(getter, setter)



"""constrained_numbers

Defines specialised property_factory's for adding constraints on numbers such
as ints and floats. The properties can enforce type, require values are within
a given range, are greater than 0, etc.

Contains:
    - floating_point
    - integer
    - positive
    - positive_float
    - positive_int
    - negative
    - negative_float
    - negative_int
    - greater_than
    - int_greater_than
    - float_greater_than
    - less_than
    - int_less_than
    - float_less_than
    - in_range
    - float_in_range
    - int_in_range
    - not_in_range
    - float_not_in_range
    - int_not_in_range
"""

from .property_factory import property_factory

# Define helper functions


def _convert_args(type_constraint=None):
    if type_constraint is None:
        return {}
    return {
        "transform": type_constraint,
        "transform_err_msg": f"Must be convertible to {type_constraint.__name__}",
    }


def _greater_than_args(x, inclusive: bool):
    return {
        "condition": (lambda y: y >= x if inclusive else y > x),
        "condition_err_msg": (
            f"Must be greater than {'or equal to ' if inclusive else ''}{x}"
        ),
    }


def _less_than_args(x, inclusive: bool):
    return {
        "condition": (lambda y: y <= x if inclusive else y < x),
        "condition_err_msg": (
            f"Must be less than {'or equal to ' if inclusive else ''}{x}"
        ),
    }


def _test_bounds(bounds):
    try:
        if len(bounds) != 2:
            raise ValueError("bounds should have len() == 2")
        if bounds[0] >= bounds[1]:
            raise ValueError("bounds[0] must be less than bounds[1]")
    except TypeError as e:
        err_msg = "bounds should be a subscriptable object with len() == 2"
        raise TypeError(err_msg) from e
    return bounds


def _ranged_inclusive_as_tuple(inclusive):
    ranged_err = (
        "When specifying if a range is inclusive, users should supply "
        "either a single bool or an subscriptable object with len()==2"
    )
    if isinstance(inclusive, bool):
        inclusive = (inclusive, inclusive)
    try:
        if len(inclusive) != 2:
            raise ValueError(ranged_err)
        inclusive[0]
        inclusive[1]
    except TypeError as e:
        raise TypeError(ranged_err) from e
    return tuple(inclusive)


def _in_range_args(bounds, inclusive: True):
    inclusive = _ranged_inclusive_as_tuple(inclusive)

    def condition(x):
        above_min = x >= bounds[0] if inclusive[0] else x > bounds[0]
        below_max = x <= bounds[1] if inclusive[1] else x < bounds[1]
        return above_min and below_max

    err_msg = (
        f"Must be within range {'[' if inclusive[0] else '('}{bounds[0]},"
        f"{bounds[1]}{']' if inclusive[1] else ')'}"
    )
    return {"condition": condition, "condition_err_msg": err_msg}


def _not_in_range_args(bounds, inclusive: False):
    inclusive = _ranged_inclusive_as_tuple(inclusive)

    def condition(x):
        below_min = x <= bounds[0] if inclusive[0] else x < bounds[0]
        above_max = x >= bounds[1] if inclusive[1] else x > bounds[1]
        return below_min or above_max

    err_msg = (
        f"Must be outside range {'[' if inclusive[0] else '('}{bounds[0]},"
        f"{bounds[1]}{']' if inclusive[1] else ')'}"
    )
    return {"condition": condition, "condition_err_msg": err_msg}


# Define properties


def floating_point():
    """Creates property that transforms to floating point"""
    return property_factory(**_convert_args(float))


def integer():
    """Creates property that transforms to int"""
    return property_factory(**_convert_args(int))


def boolean():
    """Creates property that transforms to bool"""
    return property_factory(**_convert_args(bool))


def positive(allow_zero: bool = True, type_constraint=None):
    """Creates property that must be positive, possibly including zero

    Parameters:

        allow_zero (bool): If set to False, excludes zero as a possibility.
        type_constraint: If set to None, does not constrain. Can be set to int,
            float, etc to enforce the underlying type of the property.

    Returns:

        property
    """

    return property_factory(
        **_greater_than_args(0, inclusive=allow_zero),
        **_convert_args(type_constraint),
    )


def positive_float(allow_zero: bool = True):
    """Creates property that must be positive float

    Parameters:

        allow_zero (bool): If set to False, excludes zero as a possibility.

    Returns:

        property
    """
    return positive(allow_zero=allow_zero, type_constraint=float)


def positive_int(allow_zero: bool = True):
    """Creates property that must be positive int

    Parameters:

        allow_zero (bool): If set to False, excludes zero as a possibility.

    Returns:

        property
    """
    return positive(allow_zero=allow_zero, type_constraint=int)


def negative(allow_zero: bool = False, type_constraint=None):
    """Creates property that must be negative, possibly including zero

    Parameters:

        allow_zero (bool): If set to True, includes zero as a possibility.
        type_constraint: If set to None, does not constrain. Can be set to int,
            float, etc to enforce the underlying type of the property.

    Returns:

        property
    """
    return property_factory(
        **_less_than_args(0, inclusive=allow_zero),
        **_convert_args(type_constraint),
    )


def negative_float(allow_zero: bool = True):
    """Creates property that must be negative float

    Parameters:

        allow_zero (bool): If set to True, includes zero as a possibility.

    Returns:

        property
    """
    return negative(allow_zero=allow_zero, type_constraint=float)


def negative_int(allow_zero: bool = True):
    """Creates property that must be negative int

    Parameters:

        allow_zero (bool): If set to True, includes zero as a possibility.

    Returns:

        property
    """
    return negative(allow_zero=allow_zero, type_constraint=int)


def greater_than(x, inclusive: bool = False, type_constraint=None):
    """Creates property that must be greater than (or equal to) some value

    Parameters:

        x: Value that the property must be greater than
        inclusive (bool): If set to True, includes x as a possibility.
        type_constraint: If set to None, does not constrain. Can be set to int,
            float, etc to enforce the underlying type of the property.

    Returns:

        property
    """

    return property_factory(
        **_greater_than_args(x, inclusive=inclusive),
        **_convert_args(type_constraint),
    )


def float_greater_than(x, inclusive: bool = False):
    """Creates property that must be a float greater than (or equal to)
    some value

    Parameters:

        x: Value that the property must be greater than
        inclusive (bool): If set to True, includes x as a possibility.

    Returns:

        property
    """
    return greater_than(x, inclusive=inclusive, type_constraint=float)


def int_greater_than(x, inclusive: bool = False):
    """Creates property that must be an int greater than (or equal to)
    some value

    Parameters:

        x: Value that the property must be greater than
        inclusive (bool): If set to True, includes x as a possibility.

    Returns:

        property
    """
    return greater_than(x, inclusive=inclusive, type_constraint=int)


def less_than(x, inclusive: bool = False, type_constraint=None):
    """Creates property that must be less than (or equal to) some value

    Parameters:

        x: Value that the property must be less than
        inclusive (bool): If set to True, includes x as a possibility.
        type_constraint: If set to None, does not constrain. Can be set to int,
            float, etc to enforce the underlying type of the property.

    Returns:

        property
    """

    return property_factory(
        **_less_than_args(x, inclusive=inclusive),
        **_convert_args(type_constraint),
    )


def float_less_than(x, inclusive: bool = False):
    """Creates property that must be a float less than (or equal to)
    some value

    Parameters:

        x: Value that the property must be less than
        inclusive (bool): If set to True, includes x as a possibility.

    Returns:

        property
    """
    return less_than(x, inclusive=inclusive, type_constraint=float)


def int_less_than(x, inclusive: bool = False):
    """Creates property that must be an int less than (or equal to)
    some value

    Parameters:

        x: Value that the property must be less than
        inclusive (bool): If set to True, includes x as a possibility.

    Returns:

        property
    """
    return less_than(x, inclusive=inclusive, type_constraint=int)


def in_range(bounds, inclusive=True, type_constraint=None):
    """Creates property that must be between bounds[0] and bounds[1].

    Parameters:

        bounds: Subscriptable with len()==2, where bounds[0] is the lower
            bound and bounds[1] is the upper bound.
            Requires bounds[1] > bounds[0].
        inclusive (bool): If set to False, values falling on the upper and
            lower bounds will not be accepted. Can set one bound to be
            inclusive and the other exclusive by setting this to a tuple
            of 2 bools, e.g. (True,False) makes the lower bound inclusive
            while the upper bound is not.
        type_constraint: If set to None, does not constrain. Can be set to int,
            float, etc to enforce the underlying type of the property.

    Returns:

        property
    """
    return property_factory(
        **_in_range_args(bounds, inclusive),
        **_convert_args(type_constraint),
    )


def float_in_range(bounds, inclusive=True):
    """Creates property that must be a float between bounds[0] and bounds[1].

    Parameters:

        bounds: Subscriptable with len()==2, where bounds[0] is the lower
            bound and bounds[1] is the upper bound.
            Requires bounds[1] > bounds[0].
        inclusive (bool): If set to False, values falling on the upper and
            lower bounds will not be accepted. Can set one bound to be
            inclusive and the other exclusive by setting this to a tuple
            of 2 bools, e.g. (True,False) makes the lower bound inclusive
            while the upper bound is not.

    Returns:

        property
    """
    return in_range(bounds, inclusive, type_constraint=float)


def int_in_range(bounds, inclusive=True):
    """Creates property that must be an int between bounds[0] and bounds[1].

    Parameters:

        bounds: Subscriptable with len()==2, where bounds[0] is the lower
            bound and bounds[1] is the upper bound.
            Requires bounds[1] > bounds[0].
        inclusive (bool): If set to False, values falling on the upper and
            lower bounds will not be accepted. Can set one bound to be
            inclusive and the other exclusive by setting this to a tuple
            of 2 bools, e.g. (True,False) makes the lower bound inclusive
            while the upper bound is not.

    Returns:

        property
    """
    return in_range(bounds, inclusive, type_constraint=int)


def not_in_range(bounds, inclusive=False, type_constraint=None):
    """Creates property that must be outside bounds[0] and bounds[1].

    Parameters:

        bounds: Subscriptable with len()==2, where bounds[0] is the lower
            bound and bounds[1] is the upper bound.
            Requires bounds[1] > bounds[0].
        inclusive (bool): If set to False, values falling on the upper and
            lower bounds will not be accepted. Can set one bound to be
            inclusive and the other exclusive by setting this to a tuple
            of 2 bools, e.g. (True,False) makes the lower bound inclusive
            while the upper bound is not.
        type_constraint: If set to None, does not constrain. Can be set to int,
            float, etc to enforce the underlying type of the property.

    Returns:

        property
    """
    return property_factory(
        **_not_in_range_args(bounds, inclusive),
        **_convert_args(type_constraint),
    )


def float_not_in_range(bounds, inclusive=False):
    """Creates property that must be a float outside bounds[0] and bounds[1].

    Parameters:

        bounds: Subscriptable with len()==2, where bounds[0] is the lower
            bound and bounds[1] is the upper bound.
            Requires bounds[1] > bounds[0].
        inclusive (bool): If set to False, values falling on the upper and
            lower bounds will not be accepted. Can set one bound to be
            inclusive and the other exclusive by setting this to a tuple
            of 2 bools, e.g. (True,False) makes the lower bound inclusive
            while the upper bound is not.

    Returns:

        property
    """
    return not_in_range(bounds, inclusive, type_constraint=float)


def int_not_in_range(bounds, inclusive=False):
    """Creates property that must be an int outside bounds[0] and bounds[1].

    Parameters:

        bounds: Subscriptable with len()==2, where bounds[0] is the lower
            bound and bounds[1] is the upper bound.
            Requires bounds[1] > bounds[0].
        inclusive (bool): If set to False, values falling on the upper and
            lower bounds will not be accepted. Can set one bound to be
            inclusive and the other exclusive by setting this to a tuple
            of 2 bools, e.g. (True,False) makes the lower bound inclusive
            while the upper bound is not.

    Returns:

        property
    """
    return not_in_range(bounds, inclusive, type_constraint=int)



from proper_tea import property_factory

import pytest


@pytest.fixture
def property_class():
    class MyClass:
        ends_in_dot = property_factory(
            transform=lambda z: z + ".",
            transform_err_msg="Couldn't add to end of string",
        )

        equal_to_5 = property_factory(
            condition=lambda z: z == 5,
            condition_err_msg="Should be equal to 5",
        )

        less_than_5 = property_factory(
            condition=lambda z: z < 5,
            condition_err_msg="Should be less than 5",
        )

        def __init__(self):
            self.ends_in_dot = "string"
            self.equal_to_5 = 5
            self.less_than_5 = 3

    return MyClass()


def test_property_factory_transform(property_class):
    # MyClass.ends_in_dot should always have '.' in the last place
    assert property_class.ends_in_dot[-1] == "."
    # Test reassigning repeats the transform
    property_class.ends_in_dot = "hello world"
    assert property_class.ends_in_dot == "hello world."
    # Test that it can happen recusively
    property_class.ends_in_dot = property_class.ends_in_dot
    assert property_class.ends_in_dot == "hello world.."
    # Test that it raises ValueError when passed a non string
    with pytest.raises(ValueError) as execinfo:
        property_class.ends_in_dot = 17
    assert "raise" in str(execinfo.value)
    # Ensure that a failed set doesn't modify
    assert property_class.ends_in_dot == "hello world.."


def test_property_factory_condition(property_class):
    # MyClass.equal_to_5 should fail when given anything except 5
    # Note that it should not raise an exception in the vast
    # majority of cases.
    assert property_class.equal_to_5 == 5
    assert isinstance(property_class.equal_to_5, int)
    # Test it works for floats
    property_class.equal_to_5 = 5.0
    assert property_class.equal_to_5 == 5.0
    assert isinstance(property_class.equal_to_5, float)
    # Test it returns false when given something else
    with pytest.raises(ValueError) as execinfo:
        property_class.equal_to_5 = 0
    assert "False" in str(execinfo.value)
    # Ensure that a failed set didn't change the result
    assert property_class.equal_to_5 == 5.0
    # MyClass.less_than_5 should instead raise an exception
    # if given a string. First test that it works:
    property_class.less_than_5 = 3
    assert property_class.less_than_5 == 3
    # Now try giving it a string
    with pytest.raises(ValueError) as execinfo:
        property_class.less_than_5 = "hello world."
    assert "raise" in str(execinfo.value)



import proper_tea as pt
import pytest
from itertools import product


@pytest.fixture
def basic_test():
    class MyClass:
        floating_point = pt.floating_point()
        integer = pt.integer()
        boolean = pt.boolean()

    return MyClass()


class TestBasics:
    def test_floating_point(self, basic_test):
        # Ensure the user can assign
        basic_test.floating_point = 81.3
        assert basic_test.floating_point == 81.3
        assert isinstance(basic_test.floating_point, float)
        # Ensure that they can set using int, but it remains a float
        basic_test.floating_point = 60
        assert basic_test.floating_point == 60
        assert isinstance(basic_test.floating_point, float)
        # Ensure it won't allow something that won't convert
        with pytest.raises(ValueError) as excinfo:
            basic_test.floating_point = "hello world"
        assert "convert" in str(excinfo.value)

    def test_boolean(self, basic_test):
        # Ensure the user can assign
        basic_test.boolean = True
        assert basic_test.boolean
        assert isinstance(basic_test.boolean, bool)
        basic_test.boolean = False
        assert not basic_test.boolean
        assert isinstance(basic_test.boolean, bool)
        # Anything should work, in theory
        basic_test.boolean = 60.5
        assert basic_test.boolean
        assert isinstance(basic_test.boolean, bool)
        basic_test.boolean = 0.0
        assert not basic_test.boolean
        assert isinstance(basic_test.boolean, bool)
        basic_test.boolean = "hello world"
        assert basic_test.boolean
        assert isinstance(basic_test.boolean, bool)
        basic_test.boolean = []
        assert not basic_test.boolean
        assert isinstance(basic_test.boolean, bool)
        basic_test.boolean = None
        assert not basic_test.boolean
        assert isinstance(basic_test.boolean, bool)

    def test_integer(self, basic_test):
        # Ensure the user can assign
        basic_test.integer = 27
        assert basic_test.integer == 27
        assert isinstance(basic_test.integer, int)
        # Ensure that they can set using float, but it rounds to int
        basic_test.integer = 60.5
        assert basic_test.integer == 60
        assert isinstance(basic_test.integer, int)
        # Ensure it won't allow something that won't convert
        with pytest.raises(ValueError) as excinfo:
            basic_test.integer = "hello world"
        assert "convert" in str(excinfo.value)


@pytest.fixture
def pos_neg_test(allow_zero: bool):
    class MyClass:

        pos = pt.positive(allow_zero=allow_zero)
        pos_float = pt.positive_float(allow_zero=allow_zero)
        pos_int = pt.positive_int(allow_zero=allow_zero)
        neg = pt.negative(allow_zero=allow_zero)
        neg_float = pt.negative_float(allow_zero=allow_zero)
        neg_int = pt.negative_int(allow_zero=allow_zero)

        def __init__(self):
            self.allow_zero = allow_zero

    return MyClass()


class TestPosNeg:
    @pytest.mark.parametrize("allow_zero", [False, True])
    def test_positive(self, pos_neg_test):
        # Ensure the user can assign any positive
        pos_neg_test.pos = 4
        assert pos_neg_test.pos == 4
        assert isinstance(pos_neg_test.pos, int)
        pos_neg_test.pos = 4.5
        assert pos_neg_test.pos == 4.5
        assert isinstance(pos_neg_test.pos, float)
        # Ensure negatives fail
        with pytest.raises(ValueError) as excinfo:
            pos_neg_test.pos = -4
        assert "greater than" in str(excinfo.value)
        with pytest.raises(ValueError) as excinfo:
            pos_neg_test.pos = -4.5
        assert "greater than" in str(excinfo.value)
        # Ensure 0 works if allow_zero is true
        if pos_neg_test.allow_zero:
            pos_neg_test.pos = 0
            assert pos_neg_test.pos == 0
            with pytest.raises(ValueError) as excinfo:
                pos_neg_test.pos = -4.5
            assert "or equal to" in str(excinfo.value)
        else:
            with pytest.raises(ValueError) as excinfo:
                pos_neg_test.pos = 0
            assert "or equal to" not in str(excinfo.value)
        # Ensure it won't allow non-numbers
        with pytest.raises(ValueError) as excinfo:
            pos_neg_test.pos = "hello world"

    @pytest.mark.parametrize("allow_zero", [False, True])
    def test_positive_float(self, pos_neg_test):
        # Ensure the user can assign any positive float
        pos_neg_test.pos_float = 4.5
        assert pos_neg_test.pos_float == 4.5
        assert isinstance(pos_neg_test.pos_float, float)
        # Ensure ints are converted
        pos_neg_test.pos_float = 4
        assert pos_neg_test.pos_float == 4
        assert isinstance(pos_neg_test.pos_float, float)
        # Ensure negatives fail
        with pytest.raises(ValueError) as excinfo:
            pos_neg_test.pos_float = -4.5
        assert "greater than" in str(excinfo.value)
        with pytest.raises(ValueError) as excinfo:
            pos_neg_test.pos_float = -4
        assert "greater than" in str(excinfo.value)
        # Ensure 0 works if allow_zero is true
        if pos_neg_test.allow_zero:
            pos_neg_test.pos_float = 0
            assert pos_neg_test.pos_float == 0
            with pytest.raises(ValueError) as excinfo:
                pos_neg_test.pos_float = -4.5
            assert "or equal to" in str(excinfo.value)
        else:
            with pytest.raises(ValueError) as excinfo:
                pos_neg_test.pos_float = 0
            assert "or equal to" not in str(excinfo.value)
        # Ensure it won't allow non-numbers
        with pytest.raises(ValueError) as excinfo:
            pos_neg_test.pos_float = "hello world"

    @pytest.mark.parametrize("allow_zero", [False, True])
    def test_positive_int(self, pos_neg_test):
        # Ensure the user can assign any positive int
        pos_neg_test.pos_int = 4
        assert pos_neg_test.pos_int == 4
        assert isinstance(pos_neg_test.pos_int, int)
        # Ensure floats are converted
        pos_neg_test.pos_int = 4.4
        assert pos_neg_test.pos_int == 4
        assert isinstance(pos_neg_test.pos_int, int)
        # Ensure negatives fail
        with pytest.raises(ValueError) as excinfo:
            pos_neg_test.pos_int = -4
        assert "greater than" in str(excinfo.value)
        with pytest.raises(ValueError) as excinfo:
            pos_neg_test.pos_int = -4.5
        assert "greater than" in str(excinfo.value)
        # Ensure 0 works if allow_zero is true
        if pos_neg_test.allow_zero:
            pos_neg_test.pos_int = 0
            assert pos_neg_test.pos_int == 0
            with pytest.raises(ValueError) as excinfo:
                pos_neg_test.pos_int = -4
            assert "or equal to" in str(excinfo.value)
        else:
            with pytest.raises(ValueError) as excinfo:
                pos_neg_test.pos_int = 0
            assert "or equal to" not in str(excinfo.value)
        # Ensure it won't allow non-numbers
        with pytest.raises(ValueError) as excinfo:
            pos_neg_test.pos_int = "hello world"

    @pytest.mark.parametrize("allow_zero", [False, True])
    def test_negative(self, pos_neg_test):
        # Ensure the user can assign any negative
        pos_neg_test.neg = -4
        assert pos_neg_test.neg == -4
        assert isinstance(pos_neg_test.neg, int)
        pos_neg_test.neg = -4.5
        assert pos_neg_test.neg == -4.5
        assert isinstance(pos_neg_test.neg, float)
        # Ensure positives fail
        with pytest.raises(ValueError) as excinfo:
            pos_neg_test.neg = 4
        assert "less than" in str(excinfo.value)
        with pytest.raises(ValueError) as excinfo:
            pos_neg_test.neg = 4.5
        assert "less than" in str(excinfo.value)
        # Ensure 0 works if allow_zero is true
        if pos_neg_test.allow_zero:
            pos_neg_test.neg = 0
            assert pos_neg_test.neg == 0
            with pytest.raises(ValueError) as excinfo:
                pos_neg_test.neg = 4.5
            assert "or equal to" in str(excinfo.value)
        else:
            with pytest.raises(ValueError) as excinfo:
                pos_neg_test.neg = 0
            assert "or equal to" not in str(excinfo.value)
        # Ensure it won't allow non-numbers
        with pytest.raises(ValueError) as excinfo:
            pos_neg_test.neg = "hello world"

    @pytest.mark.parametrize("allow_zero", [False, True])
    def test_negative_float(self, pos_neg_test):
        # Ensure the user can assign any negative float
        pos_neg_test.neg_float = -4.5
        assert pos_neg_test.neg_float == -4.5
        assert isinstance(pos_neg_test.neg_float, float)
        # Ensure ints are converted
        pos_neg_test.neg_float = -4
        assert pos_neg_test.neg_float == -4
        assert isinstance(pos_neg_test.neg_float, float)
        # Ensure positives fail
        with pytest.raises(ValueError) as excinfo:
            pos_neg_test.neg_float = 4.5
        assert "less than" in str(excinfo.value)
        with pytest.raises(ValueError) as excinfo:
            pos_neg_test.neg_float = 4
        assert "less than" in str(excinfo.value)
        # Ensure 0 works if allow_zero is true
        if pos_neg_test.allow_zero:
            pos_neg_test.neg_float = 0
            assert pos_neg_test.neg_float == 0
            with pytest.raises(ValueError) as excinfo:
                pos_neg_test.neg_float = 4.5
            assert "or equal to" in str(excinfo.value)
        else:
            with pytest.raises(ValueError) as excinfo:
                pos_neg_test.neg_float = 0
            assert "or equal to" not in str(excinfo.value)
        # Ensure it won't allow non-numbers
        with pytest.raises(ValueError) as excinfo:
            pos_neg_test.neg_float = "hello world"

    @pytest.mark.parametrize("allow_zero", [False, True])
    def test_negative_int(self, pos_neg_test):
        # Ensure the user can assign any negative int
        pos_neg_test.neg_int = -4
        assert pos_neg_test.neg_int == -4
        assert isinstance(pos_neg_test.neg_int, int)
        # Ensure floats are converted (should round up)
        pos_neg_test.neg_int = -4.4
        assert pos_neg_test.neg_int == -4
        assert isinstance(pos_neg_test.neg_int, int)
        # Ensure positives fail
        with pytest.raises(ValueError) as excinfo:
            pos_neg_test.neg_int = 4
        assert "less than" in str(excinfo.value)
        with pytest.raises(ValueError) as excinfo:
            pos_neg_test.neg_int = 4.5
        assert "less than" in str(excinfo.value)
        # Ensure 0 works if allow_zero is true
        if pos_neg_test.allow_zero:
            pos_neg_test.neg_int = 0
            assert pos_neg_test.neg_int == 0
            with pytest.raises(ValueError) as excinfo:
                pos_neg_test.neg_int = 4
            assert "or equal to" in str(excinfo.value)
        else:
            with pytest.raises(ValueError) as excinfo:
                pos_neg_test.neg_int = 0
            assert "or equal to" not in str(excinfo.value)
        # Ensure it won't allow non-numbers
        with pytest.raises(ValueError) as excinfo:
            pos_neg_test.neg_int = "hello world"


@pytest.fixture
def gt_lt_test(limit: float, inclusive: bool):
    class MyClass:

        gt = pt.greater_than(limit, inclusive=inclusive)
        gt_float = pt.float_greater_than(limit, inclusive=inclusive)
        gt_int = pt.int_greater_than(limit, inclusive=inclusive)
        lt = pt.less_than(limit, inclusive=inclusive)
        lt_float = pt.float_less_than(limit, inclusive=inclusive)
        lt_int = pt.int_less_than(limit, inclusive=inclusive)

        def __init__(self):
            self.limit = limit
            self.inclusive = inclusive

    return MyClass()


class TestGtLt:
    @pytest.mark.parametrize(
        "limit, inclusive",
        [(5, False), (5.5, True), (5.5, False), (5, True)],
    )
    def test_greater_than(self, gt_lt_test):
        limit = gt_lt_test.limit
        # Ensure the user can assign any number over limit
        gt_lt_test.gt = int(limit + 3)
        assert gt_lt_test.gt == int(limit + 3)
        assert isinstance(gt_lt_test.gt, int)
        gt_lt_test.gt = limit + 3.5
        assert gt_lt_test.gt == limit + 3.5
        assert isinstance(gt_lt_test.gt, float)
        # Ensure anything less than fails
        with pytest.raises(ValueError) as excinfo:
            gt_lt_test.gt = int(limit - 4)
        assert "greater than" in str(excinfo.value)
        with pytest.raises(ValueError) as excinfo:
            gt_lt_test.gt = limit - 6.2
        assert "greater than" in str(excinfo.value)
        # Ensure limit works if inclusive is true
        if gt_lt_test.inclusive:
            gt_lt_test.gt = limit
            assert gt_lt_test.gt == limit
            with pytest.raises(ValueError) as excinfo:
                gt_lt_test.gt = limit - 1
            assert "or equal to" in str(excinfo.value)
        else:
            with pytest.raises(ValueError) as excinfo:
                gt_lt_test.gt = limit
            assert "or equal to" not in str(excinfo.value)
        # Ensure it won't allow non-numbers
        with pytest.raises(ValueError) as excinfo:
            gt_lt_test.gt = "hello world"

    @pytest.mark.parametrize(
        "limit, inclusive",
        [(5, False), (5.5, True), (5.5, False), (5, True)],
    )
    def test_float_greater_than(self, gt_lt_test):
        limit = gt_lt_test.limit
        # Ensure the user can assign any float over limit
        gt_lt_test.gt_float = limit + 2.2
        assert gt_lt_test.gt_float == limit + 2.2
        assert isinstance(gt_lt_test.gt_float, float)
        # test that ints are converted
        gt_lt_test.gt_float = int(limit + 1)
        assert gt_lt_test.gt_float == int(limit + 1)
        assert isinstance(gt_lt_test.gt_float, float)
        # Ensure anything less than fails
        with pytest.raises(ValueError) as excinfo:
            gt_lt_test.gt_float = limit - 1.1
        assert "greater than" in str(excinfo.value)
        with pytest.raises(ValueError) as excinfo:
            gt_lt_test.gt_float = limit - 10
        assert "greater than" in str(excinfo.value)
        # Ensure limit works if inclusive is true
        if gt_lt_test.inclusive:
            gt_lt_test.gt_float = limit
            assert gt_lt_test.gt_float == limit
            with pytest.raises(ValueError) as excinfo:
                gt_lt_test.gt_float = limit - 1
            assert "or equal to" in str(excinfo.value)
        else:
            with pytest.raises(ValueError) as excinfo:
                gt_lt_test.gt_float = limit
            assert "or equal to" not in str(excinfo.value)
        # Ensure it won't allow non-numbers
        with pytest.raises(ValueError) as excinfo:
            gt_lt_test.gt_float = "hello world"

    @pytest.mark.parametrize(
        "limit, inclusive",
        [(5, False), (5.5, True), (5.5, False), (5, True)],
    )
    def test_int_greater_than(self, gt_lt_test):
        limit = gt_lt_test.limit
        # Ensure the user can assign any int over limit
        gt_lt_test.gt_int = limit + 7
        assert gt_lt_test.gt_int == int(limit + 7)
        assert isinstance(gt_lt_test.gt_int, int)
        # test that floats are converted
        gt_lt_test.gt_int = limit + 2.3
        assert gt_lt_test.gt_int == int(limit + 2.3)
        assert isinstance(gt_lt_test.gt_int, int)
        # Ensure anything less than fails
        with pytest.raises(ValueError) as excinfo:
            gt_lt_test.gt_int = limit - 1
        assert "greater than" in str(excinfo.value)
        with pytest.raises(ValueError) as excinfo:
            gt_lt_test.gt_int = limit - 0.1
        assert "greater than" in str(excinfo.value)
        # Ensure limit works if inclusive is true
        if gt_lt_test.inclusive:
            gt_lt_test.gt_int = limit
            assert gt_lt_test.gt_int == int(limit)
            with pytest.raises(ValueError) as excinfo:
                gt_lt_test.gt_int = limit - 1
            assert "or equal to" in str(excinfo.value)
        else:
            with pytest.raises(ValueError) as excinfo:
                gt_lt_test.gt_int = limit
            assert "or equal to" not in str(excinfo.value)
        # Ensure it won't allow non-numbers
        with pytest.raises(ValueError) as excinfo:
            gt_lt_test.gt_int = "hello world"

    @pytest.mark.parametrize(
        "limit, inclusive",
        [(5, False), (5.5, True), (5.5, False), (5, True)],
    )
    def test_less_than(self, gt_lt_test):
        limit = gt_lt_test.limit
        # Ensure the user can assign any number under limit
        gt_lt_test.lt = int(limit - 3)
        assert gt_lt_test.lt == int(limit - 3)
        assert isinstance(gt_lt_test.lt, int)
        gt_lt_test.lt = limit - 3.5
        assert gt_lt_test.lt == limit - 3.5
        assert isinstance(gt_lt_test.lt, float)
        # Ensure anything greater than fails
        with pytest.raises(ValueError) as excinfo:
            gt_lt_test.lt = int(limit + 4)
        assert "less than" in str(excinfo.value)
        with pytest.raises(ValueError) as excinfo:
            gt_lt_test.lt = limit + 6.2
        assert "less than" in str(excinfo.value)
        # Ensure limit works if inclusive is true
        if gt_lt_test.inclusive:
            gt_lt_test.lt = limit
            assert gt_lt_test.lt == limit
            with pytest.raises(ValueError) as excinfo:
                gt_lt_test.lt = limit + 1
            assert "or equal to" in str(excinfo.value)
        else:
            with pytest.raises(ValueError) as excinfo:
                gt_lt_test.lt = limit
            assert "or equal to" not in str(excinfo.value)
        # Ensure it won't allow non-numbers
        with pytest.raises(ValueError) as excinfo:
            gt_lt_test.lt = "hello world"

    @pytest.mark.parametrize(
        "limit, inclusive",
        [(5, False), (5.5, True), (5.5, False), (5, True)],
    )
    def test_float_less_than(self, gt_lt_test):
        limit = gt_lt_test.limit
        # Ensure the user can assign any float under limit
        gt_lt_test.lt_float = limit - 2.2
        assert gt_lt_test.lt_float == limit - 2.2
        assert isinstance(gt_lt_test.lt_float, float)
        # test that ints are converted
        gt_lt_test.lt_float = int(limit - 1)
        assert gt_lt_test.lt_float == int(limit - 1)
        assert isinstance(gt_lt_test.lt_float, float)
        # Ensure anything greater than fails
        with pytest.raises(ValueError) as excinfo:
            gt_lt_test.lt_float = limit + 1.1
        assert "less than" in str(excinfo.value)
        with pytest.raises(ValueError) as excinfo:
            gt_lt_test.lt_float = limit + 10
        assert "less than" in str(excinfo.value)
        # Ensure limit works if inclusive is true
        if gt_lt_test.inclusive:
            gt_lt_test.lt_float = limit
            assert gt_lt_test.lt_float == limit
            with pytest.raises(ValueError) as excinfo:
                gt_lt_test.lt_float = limit + 1
            assert "or equal to" in str(excinfo.value)
        else:
            with pytest.raises(ValueError) as excinfo:
                gt_lt_test.lt_float = limit
            assert "or equal to" not in str(excinfo.value)
        # Ensure it won't allow non-numbers
        with pytest.raises(ValueError) as excinfo:
            gt_lt_test.lt_float = "hello world"

    @pytest.mark.parametrize(
        "limit, inclusive",
        [(5, False), (5.5, True), (5.5, False), (5, True)],
    )
    def test_int_less_than(self, gt_lt_test):
        limit = gt_lt_test.limit
        # Ensure the user can assign any int under limit
        gt_lt_test.lt_int = limit - 7
        assert gt_lt_test.lt_int == int(limit - 7)
        assert isinstance(gt_lt_test.lt_int, int)
        # test that floats are converted
        gt_lt_test.lt_int = limit - 2.3
        assert gt_lt_test.lt_int == int(limit - 2.3)
        assert isinstance(gt_lt_test.lt_int, int)
        # Ensure anything greater than fails
        with pytest.raises(ValueError) as excinfo:
            gt_lt_test.lt_int = limit + 1
        assert "less than" in str(excinfo.value)
        with pytest.raises(ValueError) as excinfo:
            gt_lt_test.lt_int = limit + 0.1
        assert "less than" in str(excinfo.value)
        # Ensure limit works if inclusive is true
        if gt_lt_test.inclusive:
            gt_lt_test.lt_int = limit
            assert gt_lt_test.lt_int == int(limit)
            with pytest.raises(ValueError) as excinfo:
                gt_lt_test.lt_int = limit + 1
            assert "or equal to" in str(excinfo.value)
        else:
            with pytest.raises(ValueError) as excinfo:
                gt_lt_test.lt_int = limit
            assert "or equal to" not in str(excinfo.value)
        # Ensure it won't allow non-numbers
        with pytest.raises(ValueError) as excinfo:
            gt_lt_test.lt_int = "hello world"


@pytest.fixture
def ranged_test(bounds: (float, float), inclusive: (bool, bool)):
    class MyClass:

        in_range = pt.in_range(bounds=bounds, inclusive=inclusive)
        float_in_range = pt.float_in_range(bounds=bounds, inclusive=inclusive)
        int_in_range = pt.int_in_range(bounds=bounds, inclusive=inclusive)
        not_in_range = pt.not_in_range(bounds=bounds, inclusive=inclusive)
        float_not_in_range = pt.float_not_in_range(bounds=bounds, inclusive=inclusive)
        int_not_in_range = pt.int_not_in_range(bounds=bounds, inclusive=inclusive)

        def __init__(self):
            self.bounds = bounds
            self.inclusive = inclusive
            if isinstance(inclusive, bool):
                self.inclusive = (inclusive, inclusive)

    return MyClass()


def get_ranged_params():
    return product(
        product((-2.2, -2), (+2.2, +2)),
        [*product((False, True), (False, True)), False, True],
    )


class TestRanged:
    @pytest.mark.parametrize("bounds, inclusive", get_ranged_params())
    def test_in_range(self, ranged_test):
        bounds = ranged_test.bounds
        inclusive = ranged_test.inclusive
        # Ensure the user can assign number within bounds
        ranged_test.in_range = 0.5 * sum(bounds)
        assert ranged_test.in_range == 0.5 * sum(bounds)
        assert isinstance(ranged_test.in_range, float)
        ranged_test.in_range = int(0.5 * sum(bounds))
        assert ranged_test.in_range == int(0.5 * sum(bounds))
        assert isinstance(ranged_test.in_range, int)
        # Ensure anything outside fails
        with pytest.raises(ValueError) as excinfo:
            ranged_test.in_range = bounds[0] - 1
        assert "range" in str(excinfo.value)
        with pytest.raises(ValueError) as excinfo:
            ranged_test.in_range = bounds[1] + 1
        assert "range" in str(excinfo.value)
        # Ensure limit works if inclusive is true
        if ranged_test.inclusive[0]:
            ranged_test.in_range = bounds[0]
            assert ranged_test.in_range == bounds[0]
            with pytest.raises(ValueError) as excinfo:
                ranged_test.in_range = bounds[0] - 1
            assert "[" in str(excinfo.value)
        else:
            with pytest.raises(ValueError) as excinfo:
                ranged_test.in_range = bounds[0]
            assert "(" in str(excinfo.value)
        if ranged_test.inclusive[1]:
            ranged_test.in_range = bounds[1]
            assert ranged_test.in_range == bounds[1]
            with pytest.raises(ValueError) as excinfo:
                ranged_test.in_range = bounds[1] + 1
            assert "]" in str(excinfo.value)
        else:
            with pytest.raises(ValueError) as excinfo:
                ranged_test.in_range = bounds[1]
            assert ")" in str(excinfo.value)
        # Ensure it won't allow non-numbers
        with pytest.raises(ValueError) as excinfo:
            ranged_test.in_range = "hello world"

    @pytest.mark.parametrize("bounds, inclusive", get_ranged_params())
    def test_float_in_range(self, ranged_test):
        bounds = ranged_test.bounds
        inclusive = ranged_test.inclusive
        # Ensure the user can assign number within bounds
        ranged_test.float_in_range = 0.5 * sum(bounds)
        assert ranged_test.float_in_range == 0.5 * sum(bounds)
        assert isinstance(ranged_test.float_in_range, float)
        ranged_test.float_in_range = int(0.5 * sum(bounds))
        assert ranged_test.float_in_range == int(0.5 * sum(bounds))
        assert isinstance(ranged_test.float_in_range, float)
        # Ensure anything outside fails
        with pytest.raises(ValueError) as excinfo:
            ranged_test.float_in_range = bounds[0] - 1
        assert "range" in str(excinfo.value)
        with pytest.raises(ValueError) as excinfo:
            ranged_test.float_in_range = bounds[1] + 1
        # Ensure limit works if inclusive is true
        if ranged_test.inclusive[0]:
            ranged_test.float_in_range = bounds[0]
            assert ranged_test.float_in_range == bounds[0]
            with pytest.raises(ValueError) as excinfo:
                ranged_test.float_in_range = bounds[0] - 1
            assert "[" in str(excinfo.value)
        else:
            with pytest.raises(ValueError) as excinfo:
                ranged_test.float_in_range = bounds[0]
            assert "(" in str(excinfo.value)
        if ranged_test.inclusive[1]:
            ranged_test.float_in_range = bounds[1]
            assert ranged_test.float_in_range == bounds[1]
            with pytest.raises(ValueError) as excinfo:
                ranged_test.float_in_range = bounds[1] + 1
            assert "]" in str(excinfo.value)
        else:
            with pytest.raises(ValueError) as excinfo:
                ranged_test.float_in_range = bounds[1]
            assert ")" in str(excinfo.value)
        # Ensure it won't allow non-numbers
        with pytest.raises(ValueError) as excinfo:
            ranged_test.float_in_range = "hello world"

    @pytest.mark.parametrize("bounds, inclusive", get_ranged_params())
    def test_int_in_range(self, ranged_test):
        bounds = ranged_test.bounds
        inclusive = ranged_test.inclusive
        # Ensure the user can assign number within bounds
        ranged_test.int_in_range = 0.5 * sum(bounds)
        assert ranged_test.int_in_range == int(0.5 * sum(bounds))
        assert isinstance(ranged_test.int_in_range, int)
        # Ensure anything outside fails
        with pytest.raises(ValueError) as excinfo:
            ranged_test.int_in_range = bounds[0] - 1
        assert "range" in str(excinfo.value)
        with pytest.raises(ValueError) as excinfo:
            ranged_test.int_in_range = bounds[1] + 1
        # Ensure limit works if inclusive is true
        if ranged_test.inclusive[0]:
            ranged_test.int_in_range = bounds[0]
            assert ranged_test.int_in_range == int(bounds[0])
            with pytest.raises(ValueError) as excinfo:
                ranged_test.int_in_range = bounds[0] - 1
            assert "[" in str(excinfo.value)
        else:
            with pytest.raises(ValueError) as excinfo:
                ranged_test.int_in_range = bounds[0]
            assert "(" in str(excinfo.value)
        if ranged_test.inclusive[1]:
            ranged_test.int_in_range = bounds[1]
            assert ranged_test.int_in_range == int(bounds[1])
            with pytest.raises(ValueError) as excinfo:
                ranged_test.int_in_range = bounds[1] + 1
            assert "]" in str(excinfo.value)
        else:
            with pytest.raises(ValueError) as excinfo:
                ranged_test.int_in_range = bounds[1]
            assert ")" in str(excinfo.value)
        # Ensure it won't allow non-numbers
        with pytest.raises(ValueError) as excinfo:
            ranged_test.int_in_range = "hello world"

    @pytest.mark.parametrize("bounds, inclusive", get_ranged_params())
    def test_not_in_range(self, ranged_test):
        bounds = ranged_test.bounds
        inclusive = ranged_test.inclusive
        # Ensure the user can assign number outside bounds
        ranged_test.not_in_range = bounds[0] - 1.1
        assert ranged_test.not_in_range == bounds[0] - 1.1
        assert isinstance(ranged_test.not_in_range, float)
        ranged_test.not_in_range = bounds[1] + 1.1
        assert ranged_test.not_in_range == bounds[1] + 1.1
        assert isinstance(ranged_test.not_in_range, float)
        ranged_test.not_in_range = int(bounds[0]) - 1
        assert ranged_test.not_in_range == int(bounds[0]) - 1
        assert isinstance(ranged_test.not_in_range, int)
        ranged_test.not_in_range = int(bounds[1]) + 1
        assert ranged_test.not_in_range == int(bounds[1]) + 1
        assert isinstance(ranged_test.not_in_range, int)
        # Ensure anything inside fails
        with pytest.raises(ValueError) as excinfo:
            ranged_test.not_in_range = 0.5 * sum(bounds)
        assert "range" in str(excinfo.value)
        # Ensure limit works if inclusive is true
        if ranged_test.inclusive[0]:
            ranged_test.not_in_range = bounds[0]
            assert ranged_test.not_in_range == bounds[0]
            with pytest.raises(ValueError) as excinfo:
                ranged_test.not_in_range = bounds[0] + 1
            assert "[" in str(excinfo.value)
        else:
            with pytest.raises(ValueError) as excinfo:
                ranged_test.not_in_range = bounds[0]
            assert "(" in str(excinfo.value)
        if ranged_test.inclusive[1]:
            ranged_test.not_in_range = bounds[1]
            assert ranged_test.not_in_range == bounds[1]
            with pytest.raises(ValueError) as excinfo:
                ranged_test.not_in_range = bounds[1] - 1
            assert "]" in str(excinfo.value)
        else:
            with pytest.raises(ValueError) as excinfo:
                ranged_test.not_in_range = bounds[1]
            assert ")" in str(excinfo.value)
        # Ensure it won't allow non-numbers
        with pytest.raises(ValueError) as excinfo:
            ranged_test.not_in_range = "hello world"

    @pytest.mark.parametrize("bounds, inclusive", get_ranged_params())
    def test_float_not_in_range(self, ranged_test):
        bounds = ranged_test.bounds
        inclusive = ranged_test.inclusive
        # Ensure the user can assign number outside bounds
        ranged_test.float_not_in_range = bounds[0] - 1.1
        assert ranged_test.float_not_in_range == bounds[0] - 1.1
        assert isinstance(ranged_test.float_not_in_range, float)
        ranged_test.float_not_in_range = bounds[1] + 1.1
        assert ranged_test.float_not_in_range == bounds[1] + 1.1
        assert isinstance(ranged_test.float_not_in_range, float)
        ranged_test.float_not_in_range = int(bounds[0]) - 1
        assert ranged_test.float_not_in_range == int(bounds[0]) - 1
        assert isinstance(ranged_test.float_not_in_range, float)
        ranged_test.float_not_in_range = int(bounds[1]) + 1
        assert ranged_test.float_not_in_range == int(bounds[1]) + 1
        assert isinstance(ranged_test.float_not_in_range, float)
        # Ensure anything inside fails
        with pytest.raises(ValueError) as excinfo:
            ranged_test.float_not_in_range = 0.5 * sum(bounds)
        assert "range" in str(excinfo.value)
        # Ensure limit works if inclusive is true
        if ranged_test.inclusive[0]:
            ranged_test.float_not_in_range = bounds[0]
            assert ranged_test.float_not_in_range == bounds[0]
            with pytest.raises(ValueError) as excinfo:
                ranged_test.float_not_in_range = bounds[0] + 1
            assert "[" in str(excinfo.value)
        else:
            with pytest.raises(ValueError) as excinfo:
                ranged_test.float_not_in_range = bounds[0]
            assert "(" in str(excinfo.value)
        if ranged_test.inclusive[1]:
            ranged_test.float_not_in_range = bounds[1]
            assert ranged_test.float_not_in_range == bounds[1]
            with pytest.raises(ValueError) as excinfo:
                ranged_test.float_not_in_range = bounds[1] - 1
            assert "]" in str(excinfo.value)
        else:
            with pytest.raises(ValueError) as excinfo:
                ranged_test.float_not_in_range = bounds[1]
            assert ")" in str(excinfo.value)
        # Ensure it won't allow non-numbers
        with pytest.raises(ValueError) as excinfo:
            ranged_test.float_not_in_range = "hello world"

    @pytest.mark.parametrize("bounds, inclusive", get_ranged_params())
    def test_int_not_in_range(self, ranged_test):
        bounds = ranged_test.bounds
        inclusive = ranged_test.inclusive
        # Ensure the user can assign number outside bounds
        ranged_test.int_not_in_range = bounds[0] - 1.1
        assert ranged_test.int_not_in_range == int(bounds[0] - 1.1)
        assert isinstance(ranged_test.int_not_in_range, int)
        ranged_test.int_not_in_range = bounds[1] + 1.1
        assert ranged_test.int_not_in_range == int(bounds[1] + 1.1)
        assert isinstance(ranged_test.int_not_in_range, int)
        # Ensure anything inside fails
        with pytest.raises(ValueError) as excinfo:
            ranged_test.int_not_in_range = 0.5 * sum(bounds)
        assert "range" in str(excinfo.value)
        # Ensure limit works if inclusive is true
        if ranged_test.inclusive[0]:
            ranged_test.int_not_in_range = bounds[0]
            assert ranged_test.int_not_in_range == int(bounds[0])
            with pytest.raises(ValueError) as excinfo:
                ranged_test.int_not_in_range = bounds[0] + 1
            assert "[" in str(excinfo.value)
        else:
            with pytest.raises(ValueError) as excinfo:
                ranged_test.int_not_in_range = bounds[0]
            assert "(" in str(excinfo.value)
        if ranged_test.inclusive[1]:
            ranged_test.int_not_in_range = bounds[1]
            assert ranged_test.int_not_in_range == int(bounds[1])
            with pytest.raises(ValueError) as excinfo:
                ranged_test.int_not_in_range = bounds[1] - 1
            assert "]" in str(excinfo.value)
        else:
            with pytest.raises(ValueError) as excinfo:
                ranged_test.int_not_in_range = bounds[1]
            assert ")" in str(excinfo.value)
        # Ensure it won't allow non-numbers
        with pytest.raises(ValueError) as excinfo:
            ranged_test.int_not_in_range = "hello world"



"""numpy_arrays

Defines specialised property_factory's that convert inputs to numpy arrays. Numpy is
not a dependency of proper_tea, so users must have installed numpy separately in
order for these features to work.
"""

from ..property_factory import property_factory
import numpy as np


def numpy_array(shape=None, dtype=None, sort: bool = False):
    """Creates property that converts to numpy array.

    Uses np.asarray, so does not copy when passed a valid array.

    Parameters:

        shape: Optionally require a given shape.
        dtype: Optionally require the property to convert to given datatype.
        sort: Optionally sort any input arrays

    Returns:

        property
    """
    # if a single int is given as shape, convert to 1-tuple
    if isinstance(shape, int):
        shape = (shape,)

    def condition(x):
        if shape is not None:
            # Ensure shape of x is same as that provided by user
            return np.array_equal(np.shape(x), shape)
        # If shape is None, don't perform checks
        return True

    condition_err_msg = f"Must have shape {shape}"

    asarray = lambda x: np.asarray(x, dtype=dtype)
    if sort:
        transform = lambda x: np.sort(asarray(x))
    else:
        transform = asarray

    transform_err_msg = "Must be convertable to NumPy array"
    if dtype is not None:
        transform_err_msg += f" with datatype {dtype.__name__}"

    return property_factory(
        condition=condition,
        condition_err_msg=condition_err_msg,
        transform=transform,
        transform_err_msg=transform_err_msg,
    )



import proper_tea as pt
import pytest


@pytest.fixture
def in_set_test_class():
    class MyClass:
        alphabet = pt.in_set({"A", "B", "C"})
        numbers = pt.in_set({1, 2, 3})
        strings = pt.in_set({"foo", "bar"})

    return MyClass()


def test_in_set(in_set_test_class):
    # Try setting each where we expect it to work
    in_set_test_class.alphabet = "A"
    assert in_set_test_class.alphabet == "A"
    in_set_test_class.alphabet = "B"
    assert in_set_test_class.alphabet == "B"
    in_set_test_class.alphabet = "C"
    assert in_set_test_class.alphabet == "C"
    in_set_test_class.numbers = 1
    assert in_set_test_class.numbers == 1
    assert isinstance(in_set_test_class.numbers, int)
    in_set_test_class.numbers = 2
    assert in_set_test_class.numbers == 2
    assert isinstance(in_set_test_class.numbers, int)
    in_set_test_class.numbers = 3
    assert in_set_test_class.numbers == 3
    assert isinstance(in_set_test_class.numbers, int)
    in_set_test_class.numbers = 1.0
    assert in_set_test_class.numbers == 1
    assert isinstance(in_set_test_class.numbers, float)
    in_set_test_class.numbers = 2.0
    assert in_set_test_class.numbers == 2
    assert isinstance(in_set_test_class.numbers, float)
    in_set_test_class.numbers = 3.0
    assert in_set_test_class.numbers == 3
    assert isinstance(in_set_test_class.numbers, float)
    in_set_test_class.strings = "foo"
    assert in_set_test_class.strings == "foo"
    assert isinstance(in_set_test_class.strings, str)
    in_set_test_class.strings = "bar"
    assert in_set_test_class.strings == "bar"
    assert isinstance(in_set_test_class.strings, str)
    # Test a variety of failure cases
    alphabet_fail_cases = ["D", "hello world", 5, "a", None]
    for fail_case in alphabet_fail_cases:
        with pytest.raises(ValueError) as excinfo:
            in_set_test_class.alphabet = fail_case
    numbers_fail_cases = [4, "hello world", "1", None]
    for fail_case in numbers_fail_cases:
        with pytest.raises(ValueError) as excinfo:
            in_set_test_class.numbers = fail_case
    strings_fail_cases = ["f", "Foo", 5, None]
    for fail_case in strings_fail_cases:
        with pytest.raises(ValueError) as excinfo:
            in_set_test_class.strings = fail_case



"""discrete_sets

Defines specialised property_factory's that require properties may only be within a 
user-defined set.
"""

from .property_factory import property_factory


def in_set(discrete_set):
    """Creates property that must take its value from a given set

    Parameters:

        discrete_set (Iterable): Defines the set of values that the property may take.

    Returs:

        property
    """
    condition = lambda x: x in set(discrete_set)
    condition_err_msg = f"Must be one of {set(discrete_set)}"
    return property_factory(
        condition=condition,
        condition_err_msg=condition_err_msg,
    )
